import json
from typing import Callable, Tuple


def recurse_json(contents: str, fnc: Callable) -> Tuple[str, bool]:
    """Recurse with a callback function in JSON objects.
    Converts contents str into dict. This function is a wrapper for recurse_dict.

    Args:
        contents:
            JSON contents string
        fnc:
            Callback function for recursion.
            This function must return a tuple of (val, modified)
            where modified is Boolean
        depth:
            Recursion depth limit to prevent self reference

    Returns:
        new_contents:
            New JSON contents string
        modified:
            Whether it is modified or not while recursive localization
    """

    def recurse_dict(
        d, fnc, d_parent=None, d_parent_key=None, lst=None, lst_idx=None, modified=False
    ):
        """Recurse dict with a callback function.
        This function modifies mutable d.
        """
        if isinstance(d, dict):
            for k, v in d.items():
                modified |= recurse_dict(
                    v, fnc, d_parent=d, d_parent_key=k, modified=modified
                )
        elif isinstance(d, list):
            for i, v in enumerate(d):
                modified |= recurse_dict(v, fnc, lst=d, lst_idx=i, modified=modified)
        elif isinstance(d, str):
            assert d_parent is not None or lst is not None
            new_val, modified_ = fnc(d)
            modified |= modified_

            if modified_:
                if d_parent is not None:
                    d_parent[d_parent_key] = new_val
                elif lst is not None:
                    lst[lst_idx] = new_val
                else:
                    raise ValueError("Recursion failed.")
        return modified

    d = json.loads(contents)
    # d is modified inside of recurse_dict
    modified = recurse_dict(d, fnc)

    return json.dumps(d, indent=4), modified


def recurse_tsv(contents: str, fnc: Callable, delim: str = "\t") -> Tuple[str, bool]:
    """Recurse with a callback function in TSV contents.
    Just visit each line and look at values only.

    Args:
        contents:
            TSV/CSV contents string
        fnc:
            Callback function for recursion.
            This function must return a tuple of (val, modified)
            where modified is Boolean
    Returns:
        new_contents:
            New TSV/CSV contents string
        modified:
            Whether it is modified or not while recursive localization
    """
    modified = False
    new_contents = []
    for line in contents.split("\n"):
        new_values = []
        for v in line.split(delim):
            new_val, modified_ = fnc(v)
            modified |= modified_
            if modified_:
                new_values.append(new_val)
            else:
                new_values.append(v)
        new_contents.append(delim.join(new_values))

    return "\n".join(new_contents), modified


def recurse_csv(contents: str, fnc: Callable, delim: str = ",") -> Tuple[str, bool]:
    return recurse_tsv(contents, fnc, delim=delim)



"""Files for testing

Use v6_txt to test basic/simple methods of AutoURI.

Use j1_json to test recursive localization of a file containing file paths in it.
j1_json has TSV/CSV files, which can also be recursively localized.

See j1_json_contents() for its detailed structure.
    j1_json
        v41_json
        v421_tsv
            v5_csv
                v6_txt
            v6_txt
            v1_json (optionally to check self-reference)
        v5_csv
            v6_txt
"""
import os

from autouri.autouri import AutoURI
from autouri.loc_aux import recurse_csv, recurse_json, recurse_tsv


def j1_json_contents(
    prefix, prefix_v41=None, prefix_v421=None, prefix_v5=None, loc_suffix=""
):
    if prefix_v41 is None:
        prefix_v41 = prefix
    if prefix_v421 is None:
        prefix_v421 = prefix
    if prefix_v5 is None:
        prefix_v5 = prefix
    return """{{
        "k1": "v1",
        "k2": 1,
        "k3": 2.2,
        "k4": {{
            "k41": "{prefix_v41}/v41.json",
            "k42": {{
                "k421": "{prefix_v421}/deeper/v421{loc_suffix}.tsv"
            }},
            "k43": null
        }},
        "k5": "{prefix_v5}/even/deeper/v5{loc_suffix}.csv"
    }}
    """.format(
        prefix_v41=prefix_v41,
        prefix_v421=prefix_v421,
        prefix_v5=prefix_v5,
        loc_suffix=loc_suffix,
    )


def v41_json_contents():
    """JSON without any files defined in it.
    """
    return '{ "k1": "v1" }'


def v421_tsv_contents(
    prefix,
    prefix_v5=None,
    prefix_v1=None,
    prefix_v6=None,
    loc_suffix="",
    make_link_to_j1_json=False,
):
    """
    Args:
        make_link_to_j1_json:
            Link j1.json in this TSV file to to intentionally make a self reference
            , and hence an infinite loop while recursive localization
            j1.json -> v421.tsv -> j1.json -> ...
    """
    if prefix_v5 is None:
        prefix_v5 = prefix
    if prefix_v1 is None:
        prefix_v1 = prefix
    if prefix_v6 is None:
        prefix_v6 = prefix
    arr = [
        "k1\tv1",
        "k2\t1",
        "k3\t2.2",
        "k4\tnot/absolute/path",
        "k5\t{prefix_v5}/even/deeper/v5{loc_suffix}.csv",
        "k6\t{prefix_v6}/v6.txt",
    ]
    if make_link_to_j1_json:
        arr.append("k7\t{prefix_v1}/j1.json")
    s = "\n".join(arr)
    return s.format(
        prefix_v5=prefix_v5,
        prefix_v1=prefix_v1,
        prefix_v6=prefix_v6,
        loc_suffix=loc_suffix,
    )


def v5_csv_contents(prefix, prefix_v6=None):
    if prefix_v6 is None:
        prefix_v6 = prefix
    s = "\n".join(
        [
            "k1,v1",
            "k2,1",
            "k3,2.2",
            "k4,not/absolute/path",
            "k5,s33://not-valid-bucket-address",
            "k6,{prefix_v6}/v6.txt",
        ]
    )
    return s.format(prefix_v6=prefix_v6)


def v6_txt_contents():
    return "v6: Hello World"


def j1_json(
    prefix, prefix_v41=None, prefix_v421=None, prefix_v5=None, loc_suffix="", make=False
):
    u = "{prefix}/j1{loc_suffix}.json".format(prefix=prefix, loc_suffix=loc_suffix)
    if make:
        AutoURI(u).write(
            j1_json_contents(
                prefix=prefix,
                prefix_v41=prefix_v41,
                prefix_v421=prefix_v421,
                prefix_v5=prefix_v5,
                loc_suffix=loc_suffix,
            )
        )
    return u


def v41_json(prefix, make=False):
    u = "{prefix}/v41.json".format(prefix=prefix)
    if make:
        AutoURI(u).write(v41_json_contents())
    return u


def v421_tsv(
    prefix,
    prefix_v5=None,
    prefix_v1=None,
    prefix_v6=None,
    loc_suffix="",
    make=False,
    make_link_to_j1_json=False,
):
    u = "{prefix}/deeper/v421{loc_suffix}.tsv".format(
        prefix=prefix, loc_suffix=loc_suffix
    )
    if make:
        AutoURI(u).write(
            v421_tsv_contents(
                prefix=prefix,
                prefix_v5=prefix_v5,
                prefix_v1=prefix_v1,
                prefix_v6=prefix_v6,
                loc_suffix=loc_suffix,
                make_link_to_j1_json=make_link_to_j1_json,
            )
        )
    return u


def v5_csv(prefix, prefix_v6=None, make=False):
    u = "{prefix}/even/deeper/v5.csv".format(prefix=prefix)
    if make:
        AutoURI(u).write(v5_csv_contents(prefix=prefix, prefix_v6=prefix_v6))
    return u


def v6_txt(prefix, make=False):
    u = "{prefix}/v6.txt".format(prefix=prefix)
    if make:
        AutoURI(u).write(v6_txt_contents())
    return u


def common_paths():
    return [
        "/testing/abspath",
        "/testing/abspath/",
        "~/os/expandable",
        "~~/os/expandable",
        "~/~/os/expandable",
        "test/ok/man.csv",
        "http://hello.world.com/ok.txt",
        "https://hello.world.com/ok.txt",
        "http://hello.world.com/ok.txt?parameter1=true&parameter2=false",
        "https://hello.world.com/ok.txt?parameter1=true&parameter2=false",
        "http:/hello.world.com/notok.txt",
        "ftp:/hello.world.com/notok.txt",
        "dx://dnanexus-prj/not/supported.txt",
        "file:/notok.txt",
        "file://hostname/notok.txt",
        "s3://hello-world/ok.txt",
        "s3:/hello-world/not-ok.txt",
        "s3:\\hello-world\\not-ok.txt",
        "gs://hello-world/ok.txt",
        "gs:/hello-world/not-ok.txt",
        "gs:\\hello-world\\not-ok.txt",
        "!@#:;$@!#$F",
    ]


def recurse_raise_if_uri_not_exist(uri):
    uri = AutoURI(uri)
    if uri.is_valid:
        if uri.exists:
            if uri.ext == ".json":
                recurse_json(uri.read(), recurse_raise_if_uri_not_exist)
            elif uri.ext == ".tsv":
                recurse_tsv(uri.read(), recurse_raise_if_uri_not_exist)
            elif uri.ext == ".csv":
                recurse_csv(uri.read(), recurse_raise_if_uri_not_exist)
        else:
            raise Exception("URI is a valid path but does not exist.")
    return None, False


def make_files_in_dir(prefix, make_local_empty_dir_d_a=False):
    """Make a compllicated directory structure with empty files.

    Directory structure and empty files in it:
        $prefix (as root)
            a
            b/
                a
                b
            c/
                a/
                    a
                    b
                b/
                    a/
                        a
                c
            d/ (optional if make_local_empty_dir_d_a)
                a/ (optional if make_local_empty_dir_d_a)
    Args:
        make_local_empty_dir_d_a:
            Make a local empty dir ($prefix/d/a/).
            This flag should not be used for cloud buckets since
            they don't support sub-directories.
    Returns:
        List of file URIs.
    """
    file_a = os.path.join(prefix, "a")
    file_b_a = os.path.join(prefix, "b/a")
    file_b_b = os.path.join(prefix, "b/b")
    file_c_a_a = os.path.join(prefix, "c/a/a")
    file_c_a_b = os.path.join(prefix, "c/a/b")
    file_c_b_a_a = os.path.join(prefix, "c/b/a/a")
    file_c_c = os.path.join(prefix, "c/c")

    all_files = [
        file_a,
        file_b_a,
        file_b_b,
        file_c_a_a,
        file_c_a_b,
        file_c_b_a_a,
        file_c_c,
    ]

    for uri in all_files:
        AutoURI(uri).write("")
    if make_local_empty_dir_d_a:
        path = os.path.join(prefix, "d/a")
        os.makedirs(path, exist_ok=True)

    return all_files



"""GCS Bucket configuration:
    Object versioning must be turned off
        Check it with "gsutil versioning get gs://BUCKET-NAME"
        https://cloud.google.com/storage/docs/object-versioning
"""
import json
import logging
import os
import time
from datetime import timedelta
from subprocess import check_call
from tempfile import NamedTemporaryFile, TemporaryDirectory
from typing import Optional, Tuple

import requests
from filelock import BaseFileLock
from google.api_core.exceptions import (
    Forbidden,
    GatewayTimeout,
    NotFound,
    PermissionDenied,
    ServiceUnavailable,
)
from google.auth.exceptions import DefaultCredentialsError
from google.cloud import storage
from google.cloud.storage import Blob
from google.oauth2.service_account import Credentials

from .autouri import AutoURI, URIBase
from .metadata import URIMetadata, get_seconds_from_epoch, parse_md5_str

logger = logging.getLogger(__name__)


ENV_VAR_GOOGLE_APPLICATION_CREDENTIALS = "GOOGLE_APPLICATION_CREDENTIALS"
GCS_TEMPORARY_HOLD_ERROR_MSG = "is under active Temporary hold"


def add_google_app_creds_to_env(service_account_key_file):
    """Google auth with a service account.
    To globally use the key file for all GCS Client() with different thread IDs,
    update environment variable `GOOGLE_APPLICATION_CREDENTIALS` with a given
    service account key JSON file.
    """
    service_account_key_file = os.path.abspath(
        os.path.expanduser(service_account_key_file)
    )
    if ENV_VAR_GOOGLE_APPLICATION_CREDENTIALS in os.environ:
        existing_key_file = os.environ[ENV_VAR_GOOGLE_APPLICATION_CREDENTIALS]
        if not os.path.samefile(existing_key_file, service_account_key_file):
            logger.warning(
                "Environment variable {env_var} does not match with "
                "given service_account_key_file. "
                "Using application default credentials? ".format(
                    env_var=ENV_VAR_GOOGLE_APPLICATION_CREDENTIALS
                )
            )
    logger.debug(
        "Adding service account key JSON {key_file} to "
        "environment variable {env_var}".format(
            key_file=service_account_key_file,
            env_var=ENV_VAR_GOOGLE_APPLICATION_CREDENTIALS,
        )
    )
    os.environ[ENV_VAR_GOOGLE_APPLICATION_CREDENTIALS] = service_account_key_file


class GCSURILock(BaseFileLock):
    """Slow but stable locking with using GCS temporary_hold
    Hold the lock file instead of the target file that lock file protects.

    Class constants:
        DEFAULT_RETRY_RELEASE:
            Retry if release (deletion) of a lock file fails.
        DEFAULT_RETRY_RELEASE_INTERVAL:
            Interval for retrial in seconds.
    """

    DEFAULT_RETRY_RELEASE = 3
    DEFAULT_RETRY_RELEASE_INTERVAL = 3

    def __init__(
        self,
        lock_file,
        thread_id=-1,
        timeout=900,
        poll_interval=10.0,
        retry_release=DEFAULT_RETRY_RELEASE,
        retry_release_interval=DEFAULT_RETRY_RELEASE_INTERVAL,
        no_lock=False,
    ):
        super().__init__(lock_file, timeout=timeout)
        self._poll_interval = poll_interval
        self._thread_id = thread_id
        self._retry_release = retry_release
        self._retry_release_interval = retry_release_interval

    def acquire(self, timeout=None, poll_intervall=5.0):
        """Use self._poll_interval instead of poll_intervall in args
        """
        super().acquire(timeout=timeout, poll_intervall=self._poll_interval)

    def _acquire(self):
        """Try to acquire a lock.
        Once successfully acquired, lock the .lock file temporarily by setting
        blob.temporary_hold as True (similary to `gsutil retention temp set URI`).
        This will be released in self._release().

        Parse Forbidden error to check if it's raised from temporary hold.
        It can also be raised from lack of write permission, which should be re-raised.
        """
        u = GCSURI(self._lock_file, thread_id=self._thread_id)
        try:
            blob, _ = u.get_blob(new=True)
            blob.upload_from_string("")
            blob.temporary_hold = True
            blob.patch()
            self._lock_file_fd = id(self)

        except Forbidden as e:
            err_msg = json.loads(e._response._content)["error"]["message"]
            if GCS_TEMPORARY_HOLD_ERROR_MSG not in err_msg:
                raise
            logger.debug(
                "Failed to acquire a file lock. "
                "It's already locked by another process. "
                "You need to wait until it's released. "
                "Retrying until timeout. "
            )

        except (GatewayTimeout, NotFound, ServiceUnavailable) as e:
            logger.debug(
                "Failed to acquire a file lock. "
                "Server is unavailable or busy? "
                "Or too many requests? "
                "Retrying until timeout. "
                "{err}".format(err=str(e))
            )

    def _release(self):
        u = GCSURI(self._lock_file, thread_id=self._thread_id)
        for retry in range(self._retry_release):
            try:
                blob, _ = u.get_blob()
                blob.temporary_hold = False
                blob.patch()
                blob.delete()
                self._lock_file_fd = None
                break
            except Exception as e:
                error_msg = "{err}. Failed to delete a lock file: file={file}. "
                if retry == self._retry_release - 1:
                    error_msg += (
                        "You may need to manually delete a lock file. "
                        'Use "gsutil retention temp release {file}" to unlock it first. '
                        'Then use "gsutil rm -f {file}" to delete it. '
                        "Deleting a lock file itself does not affect "
                        "the file protected by it."
                    )
                error_msg = error_msg.format(err=e, file=self._lock_file)

                logger.error(error_msg)

            time.sleep(self._retry_release_interval)


class GCSURI(URIBase):
    """
    Class constants:
        LOC_PREFIX (inherited):
            Path prefix for localization. Inherited from URIBase class.
        PRIVATE_KEY_FILE:
            Path for private key file used to get presigned URLs
        DURATION_PRESIGNED_URL:
            Duration for presigned URLs in seconds.
        RETRY_BUCKET:
            Number of retrial to access a bucket.
        RETRY_BUCKET_DELAY:
            Delay for each retrial in seconds.
        USE_GSUTIL_FOR_S3 (experimental):
            This is only for direct transfer between S3 and GCS buckets.
            WARNING:
                gsutil must be configured correctly to have all
                AWS credentials in ~/.boto file.
                Run "aws configure" first and then
                run "gsutil config" to generate corrensponding ~/.boto file.

    Protected class constants:
        _CACHED_GCS_CLIENTS:
            Per-thread GCS client object is required since
            GCS client is not thread-safe.
        _CACHED_PRESIGNED_URLS:
            Can use cached presigned URLs.
        _GCS_PUBLIC_URL_FORMAT:
            End point for a bucket with public access + key path
    """

    PRIVATE_KEY_FILE: str = ""
    DURATION_PRESIGNED_URL: int = 4233600

    RETRY_BUCKET: int = 3
    RETRY_BUCKET_DELAY: int = 1
    USE_GSUTIL_FOR_S3: bool = False

    _CACHED_GCS_CLIENTS = {}
    _CACHED_GCS_ANONYMOUS_CLIENTS = {}
    _CACHED_PRESIGNED_URLS = {}
    _GCS_PUBLIC_URL_FORMAT = "http://storage.googleapis.com/{bucket}/{path}"

    _LOC_SUFFIX = ".gcs"
    _SCHEMES = ("gs://",)

    def __init__(self, uri, thread_id=-1):
        super().__init__(uri, thread_id=thread_id)

    def _get_lock(self, timeout=None, poll_interval=None):
        if timeout is None:
            timeout = GCSURI.LOCK_TIMEOUT
        if poll_interval is None:
            poll_interval = GCSURI.LOCK_POLL_INTERVAL
        return GCSURILock(
            self._uri + GCSURI.LOCK_FILE_EXT,
            thread_id=self._thread_id,
            timeout=timeout,
            poll_interval=poll_interval,
        )

    def get_metadata(self, skip_md5=False, make_md5_file=False):
        exists, mt, sz, md5 = False, None, None, None

        try:
            b, _ = self.get_blob()
            # make keys lower-case
            headers = {k.lower(): v for k, v in b._properties.items()}
            exists = True

            if not skip_md5:
                if "md5hash" in headers:
                    md5 = parse_md5_str(headers["md5hash"])
                elif "etag" in headers:
                    md5 = parse_md5_str(headers["etag"])
                if md5 is None:
                    # make_md5_file is ignored for GCSURI
                    md5 = self.md5_from_file

            if "size" in headers:
                sz = int(headers["size"])

            if "updated" in headers:
                mt = get_seconds_from_epoch(headers["updated"])
            elif "timecreated" in headers:
                mt = get_seconds_from_epoch(headers["timecreated"])

        except Exception:
            logger.debug("Failed to get metadata from {uri}".format(uri=self._uri))

        return URIMetadata(exists=exists, mtime=mt, size=sz, md5=md5)

    def read(self, byte=False):
        blob, _ = self.get_blob()
        b = blob.download_as_string()
        if byte:
            return b
        return b.decode()

    def find_all_files(self):
        cl = GCSURI.get_gcs_client(self._thread_id)
        bucket, path = self.get_bucket_path()
        sep = GCSURI.get_path_sep()
        if path:
            path = path.rstrip(sep) + sep

        result = []
        blobs = cl.list_blobs(bucket, prefix=path)
        if blobs:
            for blob in blobs:
                scheme = GCSURI.get_schemes()[0]
                uri = scheme + sep.join([bucket, blob.name])
                result.append(uri)
        return result

    def _write(self, s):
        blob, _ = self.get_blob(new=True)
        blob.upload_from_string(s)
        # blob.update()
        return

    def _rm(self):
        blob, _ = self.get_blob()
        blob.delete()
        return

    def _cp(self, dest_uri):
        """Copy from GCSURI to
            GCSURI
            S3URI: can use gsutil for direct transfer if USE_GSUTIL_FOR_S3 == True
            AbsPath
        """
        from .s3uri import S3URI
        from .abspath import AbsPath

        dest_uri = AutoURI(dest_uri)

        if isinstance(dest_uri, (GCSURI, AbsPath)):
            src_blob, src_bucket = self.get_blob()

            if src_blob is None:
                raise ValueError("Blob does not exist for {f}".format(f=self._uri))

            if isinstance(dest_uri, GCSURI):
                _, dest_path = dest_uri.get_bucket_path()
                _, dest_bucket = dest_uri.get_blob(new=True)
                src_bucket.copy_blob(src_blob, dest_bucket, dest_path)
                return True

            elif isinstance(dest_uri, AbsPath):
                dest_uri.mkdir_dirname()
                # mtime is not updated without update().
                src_blob.update()
                src_blob.download_to_filename(dest_uri._uri)
                return True

        elif isinstance(dest_uri, S3URI):
            if GCSURI.USE_GSUTIL_FOR_S3:
                rc = check_call(["gsutil", "-q", "cp", self._uri, dest_uri._uri])
                return rc == 0
            else:
                # use local temporary file instead
                with TemporaryDirectory() as tmp_d:
                    dest_uri_local = AbsPath(os.path.join(tmp_d, self.basename))
                    # lockless copy
                    self.cp(dest_uri=dest_uri_local, no_lock=True, no_checksum=True)
                    dest_uri_local.cp(dest_uri=dest_uri, no_lock=True, no_checksum=True)
                return True

        return False

    def _cp_from(self, src_uri):
        """Copy to GCSURI from
            S3URI: can use gsutil for direct transfer if USE_GSUTIL_FOR_S3 == True
            AbsPath
            HTTPURL
        """
        from .s3uri import S3URI
        from .abspath import AbsPath
        from .httpurl import HTTPURL

        src_uri = AutoURI(src_uri)

        if isinstance(src_uri, AbsPath):
            blob, _ = self.get_blob(new=True)
            blob.upload_from_filename(src_uri._uri)
            return True

        elif isinstance(src_uri, S3URI):
            if GCSURI.USE_GSUTIL_FOR_S3:
                rc = check_call(["gsutil", "-q", "cp", src_uri._uri, self._uri])
                return rc == 0
            else:
                # use local temporary file instead
                with TemporaryDirectory() as tmp_d:
                    dest_uri_local = AbsPath(os.path.join(tmp_d, self.basename))
                    src_uri.cp(dest_uri=dest_uri_local, no_lock=True, no_checksum=True)
                    dest_uri_local.cp(dest_uri=self, no_lock=True, no_checksum=True)
                return True

        elif isinstance(src_uri, HTTPURL):
            r = requests.get(
                src_uri._uri,
                stream=True,
                allow_redirects=True,
                headers=requests.utils.default_headers(),
            )
            r.raise_for_status()
            with NamedTemporaryFile() as fp:
                for chunk in r.iter_content(HTTPURL.get_http_chunk_size()):
                    fp.write(chunk)
                fp.seek(0)
                blob, _ = self.get_blob(new=True)
                blob.upload_from_file(fp)
            return True
        return False

    def get_blob(self, new=False) -> Blob:
        """GCS Client() has a bug that shows an outdated version of a file
        when using Blob() without update().
        For read-only functions (e.g. read()), need to directly call
        cl.get_bucket(bucket).get_blob(path) instead of using Blob() class.

        Also, GCS Client() is not thread-safe and it fails for a variety of reasons.
        Retry several times for whatever reasons.

        Client.get_bucket() and Client.get_bucket().get_blob() can fail
        even if the bucket is public (Storage Reader permission for allUsers
        or allAuthenticatedUsers).
        Needs an anonymous client (Client.create_anonymous_client()) for public buckets.
        If this error occurs, then retry with an anonymous client.

        Returns:
            blob:
                Blob object
            bucket_obj:
                Bucket object
        """
        bucket, path = self.get_bucket_path()
        cl = GCSURI.get_gcs_client(self._thread_id)

        bucket_obj = None
        blob = None
        for retry in range(GCSURI.RETRY_BUCKET):
            try:
                bucket_obj = cl.get_bucket(bucket)
                blob = bucket_obj.get_blob(path)
                if new and blob is None:
                    blob = Blob(name=path, bucket=bucket_obj)
                break
            except Forbidden:
                logger.debug(
                    "Bucket/blob is forbidden. Trying again with anonymous client."
                )
                cl = GCSURI.get_gcs_anonymous_client(self._thread_id)
            except NotFound:
                raise
            except PermissionDenied:
                raise
            except Exception:
                time.sleep(GCSURI.RETRY_BUCKET_DELAY)
        if blob is None:
            raise ValueError(
                "GCS blob does not exist. lack of {access_type} permission? {uri}".format(
                    access_type="write" if new else "read", uri=self._uri
                )
            )

        return blob, bucket_obj

    def get_bucket_path(self) -> Tuple[str, str]:
        """Returns a tuple of URI's S3 bucket and path.
        """
        arr = self.uri_wo_scheme.split(GCSURI.get_path_sep(), maxsplit=1)
        if len(arr) == 1:
            # root directory without path (key)
            bucket, path = arr[0], ""
        else:
            bucket, path = arr
        return bucket, path

    def get_presigned_url(
        self, duration=None, private_key_file=None, use_cached=False
    ) -> str:
        """
        Args:
            duration: Duration in seconds. This is ignored if use_cached is on.
            use_cached: Use a cached URL.
        """
        cache = GCSURI._CACHED_PRESIGNED_URLS
        if use_cached:
            if cache is not None and self._uri in cache:
                return cache[self._uri]
        # if not self.exists:
        #     raise Exception('File does not exist. f={f}'.format(self._uri))
        if private_key_file is None:
            private_key_file = os.path.expanduser(GCSURI.PRIVATE_KEY_FILE)
        else:
            private_key_file = os.path.expanduser(private_key_file)
        if not os.path.exists(private_key_file):
            raise Exception(
                "GCS private key file not found. f:{f}".format(f=private_key_file)
            )
        credentials = Credentials.from_service_account_file(private_key_file)
        duration = duration if duration is not None else GCSURI.DURATION_PRESIGNED_URL
        blob, _ = self.get_blob()
        if blob is None:
            raise ValueError("Blob does not exist for {f}".format(f=self._uri))
        url = blob.generate_signed_url(
            expiration=timedelta(seconds=duration), credentials=credentials
        )
        cache[self._uri] = url
        return url

    def get_public_url(self) -> str:
        bucket, path = self.get_bucket_path()
        return GCSURI._GCS_PUBLIC_URL_FORMAT.format(bucket=bucket, path=path)

    @staticmethod
    def get_gcs_client(thread_id) -> storage.Client:
        """Get GCS client per thread_id.

        Get default credentials (internally calling google.auth.default()) from:
            - Environment variable GOOGLE_APPLICATION_CREDENTIALS
              - Set a service account key JSON file path as this environment variable.
            - JSON file ~/.config/gcloud/application_default_credentials.json
              - To use end-user's credentials.
              - This file is created by `gcloud auth application-default login`.

        If default credentials are not found, then
            - Make/return an anonymous client instead
            - For this thread_id, it will cache anonymous client instead of failed
              client with credentials.
        """
        cl = GCSURI._CACHED_GCS_CLIENTS.get(thread_id)

        if cl is None:
            try:
                logger.debug("New GCS client for thread {id}.".format(id=thread_id))
                cl = storage.Client()
            except DefaultCredentialsError:
                cl = GCSURI.get_gcs_anonymous_client(thread_id)
            # anonymous client can also be cached here
            GCSURI._CACHED_GCS_CLIENTS[thread_id] = cl

        return cl

    @staticmethod
    def get_gcs_anonymous_client(thread_id) -> storage.Client:
        """Get GCS anonymous client per thread_id.
        """
        cl = GCSURI._CACHED_GCS_ANONYMOUS_CLIENTS.get(thread_id)

        if cl is None:
            logger.debug(
                "New GCS anonymous client for thread {id}.".format(id=thread_id)
            )
            cl = storage.Client.create_anonymous_client()
            GCSURI._CACHED_GCS_ANONYMOUS_CLIENTS[thread_id] = cl

        return cl

    @staticmethod
    def init_gcsuri(
        loc_prefix: Optional[str] = None,
        private_key_file: Optional[str] = None,
        duration_presigned_url: Optional[int] = None,
        retry_bucket: Optional[int] = None,
        retry_bucket_delay: Optional[int] = None,
        use_gsutil_for_s3: Optional[bool] = None,
    ):
        if loc_prefix is not None:
            GCSURI.LOC_PREFIX = loc_prefix
        if private_key_file is not None:
            GCSURI.PRIVATE_KEY_FILE = private_key_file
        if duration_presigned_url is not None:
            GCSURI.DURATION_PRESIGNED_URL = duration_presigned_url
        if retry_bucket is not None:
            GCSURI.RETRY_BUCKET = retry_bucket
        if retry_bucket_delay is not None:
            GCSURI.RETRY_BUCKET_DELAY = retry_bucket_delay
        if use_gsutil_for_s3 is not None:
            GCSURI.USE_GSUTIL_FOR_S3 = use_gsutil_for_s3



import errno
import glob
import hashlib
import logging
import os
import shutil
from shutil import SameFileError, copyfile
from typing import Dict, Optional

from filelock import SoftFileLock

from .autouri import AutoURI, URIBase
from .metadata import URIMetadata

logger = logging.getLogger(__name__)


class AbsPath(URIBase):
    """
    Class constants:
        LOC_PREFIX (inherited):
            Path prefix for localization. Inherited from URIBase class.
        MAP_PATH_TO_URL:
            Dict to replace path prefix with URL prefix.
            Useful to convert absolute path into URL on a web server.
        MD5_CALC_CHUNK_SIZE:
            Chunk size to calculate md5 hash of a local file.
    """

    MAP_PATH_TO_URL: Dict[str, str] = dict()
    MD5_CALC_CHUNK_SIZE: int = 4096

    _LOC_SUFFIX = ".local"
    _PATH_SEP = os.sep

    def __init__(self, uri, thread_id=-1):
        if isinstance(uri, str):
            uri = os.path.expanduser(uri)
        super().__init__(uri, thread_id=thread_id)

    @property
    def is_valid(self):
        return os.path.isabs(self._uri)

    def rmdir(self, dry_run=False, no_lock=False):
        """Do `rm -rf` instead of deleting individual files.
        For dry-run mode, call base class' method to show files to be deleted.
        """
        if not os.path.exists(self._uri):
            raise FileNotFoundError(
                "Directory does not exist. deleted already? {dir}".format(dir=self._uri)
            )
        if dry_run:
            super().rmdir(dry_run=True, no_lock=no_lock)
        else:
            shutil.rmtree(self._uri)

    def _get_lock(self, timeout=None, poll_interval=None):
        """Use filelock.SoftFileLock for AbsPath.
        filelock.SoftFileLock watches a .lock file with faster polling.
        It's stable and also platform-independent

        Args:
            poll_interval:
                This is dummy.
                Fixed polling rate defined in BaseFileLock is used.
        """
        if timeout is None:
            timeout = AbsPath.LOCK_TIMEOUT
        # create directory and use default poll_interval
        u_lock = AutoURI(self._uri + AbsPath.LOCK_FILE_EXT)
        u_lock.mkdir_dirname()
        return SoftFileLock(u_lock._uri, timeout=timeout)

    def get_metadata(self, skip_md5=False, make_md5_file=False):
        """If md5 file doesn't exist then use hashlib.md5() to calculate md5 hash
        """
        exists = os.path.exists(self._uri)
        mt, sz, md5 = None, None, None
        if exists:
            mt = os.path.getmtime(self._uri)
            sz = os.path.getsize(self._uri)
            if not skip_md5:
                md5 = self.md5_from_file
                if md5 is None:
                    md5 = self.__calc_md5sum()
                if make_md5_file:
                    self.md5_file_uri.write(md5)

        return URIMetadata(exists=exists, mtime=mt, size=sz, md5=md5)

    def read(self, byte=False):
        if byte:
            param = "rb"
        else:
            param = "r"
        with open(self._uri, param) as fp:
            return fp.read()

    def find_all_files(self):
        query = os.path.join(self._uri, "**")
        result = []
        for f in glob.glob(query, recursive=True):
            if os.path.isfile(f):
                result.append(os.path.abspath(f))
        return result

    def _write(self, s):
        self.mkdir_dirname()
        if isinstance(s, str):
            param = "w"
        else:
            param = "wb"
        with open(self._uri, param) as fp:
            fp.write(s)
        return

    def _rm(self):
        return os.remove(self._uri)

    def _cp(self, dest_uri):
        """Copy from AbsPath to other classes
        """
        dest_uri = AutoURI(dest_uri)

        if isinstance(dest_uri, AbsPath):
            dest_uri.mkdir_dirname()
            try:
                copyfile(self._uri, dest_uri._uri, follow_symlinks=True)
            except SameFileError:
                logger.debug(
                    "cp: ignored SameFileError. src={src}, dest={dest}".format(
                        src=self._uri, dest=dest_uri._uri
                    )
                )
                if os.path.islink(dest_uri._uri):
                    dest_uri._rm()
                    copyfile(self._uri, dest_uri._uri, follow_symlinks=True)

            return True
        return False

    def _cp_from(self, src_uri):
        return False

    def get_mapped_url(self, map_path_to_url=None) -> Optional[str]:
        """
        Args:
            map_path_to_url:
                dict with k, v where k is a path prefix and v is a URL prefix
                k will be replaced with v.
                If not given, defaults to use class constant AbsPath.MAP_PATH_TO_URL
        """
        if map_path_to_url is None:
            map_path_to_url = AbsPath.MAP_PATH_TO_URL
        for k, v in map_path_to_url.items():
            if k and self._uri.startswith(k):
                return self._uri.replace(k, v, 1)
        return None

    def mkdir_dirname(self):
        """Create a directory but raise if no write permission on it
        """
        os.makedirs(self.dirname, exist_ok=True)
        if not os.access(self.dirname, os.W_OK):
            raise PermissionError(
                "No permission to write on directory: {d}".format(d=self.dirname)
            )
        return

    def soft_link(self, target, force=False):
        """Make a soft link of self on target absolute path.
        If target already exists delete it and create a link.

        Args:
            target:
                Target file's absolute path or URI object.
            force:
                Delete target file (or link) if it exists
        """
        target = AbsPath(target)
        if not target.is_valid:
            raise ValueError(
                "Target path is not a valid abs path: {t}.".format(t=target.uri)
            )
        try:
            target.mkdir_dirname()
            os.symlink(self._uri, target._uri)
        except OSError as e:
            if e.errno == errno.EEXIST and force:
                target.rm()
                os.symlink(self._uri, target._uri)
            else:
                raise e

    def __calc_md5sum(self):
        """Expensive md5 calculation
        """
        logger.debug(
            "calculating md5sum hash of local file: {file}".format(file=self._uri)
        )
        hash_md5 = hashlib.md5()
        with open(self._uri, "rb") as fp:
            for chunk in iter(lambda: fp.read(AbsPath.MD5_CALC_CHUNK_SIZE), b""):
                hash_md5.update(chunk)

        logger.debug(
            "calculating md5sum is done for local file: {file}".format(file=self._uri)
        )
        return hash_md5.hexdigest()

    @staticmethod
    def get_abspath_if_exists(path):
        if isinstance(path, URIBase):
            path = path._uri
        if isinstance(path, str):
            if os.path.exists(os.path.expanduser(path)):
                return os.path.abspath(os.path.expanduser(path))
        return path

    @staticmethod
    def init_abspath(
        loc_prefix: Optional[str] = None,
        map_path_to_url: Optional[Dict[str, str]] = None,
        md5_calc_chunk_size: Optional[int] = None,
    ):
        if loc_prefix is not None:
            AbsPath.LOC_PREFIX = loc_prefix
        if map_path_to_url is not None:
            AbsPath.MAP_PATH_TO_URL = map_path_to_url
        if md5_calc_chunk_size is not None:
            AbsPath.MD5_CALC_CHUNK_SIZE = md5_calc_chunk_size



"""URIMetadata and helper functions for metadata
"""
import warnings
from base64 import b64decode
from binascii import hexlify
from collections import namedtuple
from datetime import datetime, timezone

from dateparser import parse as dateparser_parse
from dateutil.parser import parse as dateutil_parse

URIMetadata = namedtuple("URIMetadata", ("exists", "mtime", "size", "md5"))


def get_seconds_from_epoch(timestamp: str) -> float:
    """If dateutil.parser.parse cannot parse DST timezones
    (e.g. PDT, EDT) correctly, then use dateparser.parse instead.
    """
    utc_epoch = datetime(1970, 1, 1, tzinfo=timezone.utc)
    utc_t = None
    try:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            utc_t = dateutil_parse(timestamp)
    except Exception:
        pass
    if utc_t is None or utc_t.tzname() not in ("UTC", "Z"):
        utc_t = dateparser_parse(timestamp)
    utc_t = utc_t.astimezone(timezone.utc)
    return (utc_t - utc_epoch).total_seconds()


def base64_to_hex(b: str) -> str:
    return hexlify(b64decode(b)).decode()


def parse_md5_str(raw: str) -> str:
    """Check if it's based on base64 then convert it to hexadecimal string.
    """
    raw = raw.strip("\"'")
    if len(raw) == 32:
        return raw
    else:
        try:
            return base64_to_hex(raw)
        except Exception:
            pass



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="autouri",
    version="0.2.4",
    python_requires=">=3.6",
    scripts=["bin/autouri"],
    author="Jin wook Lee",
    author_email="leepc12@gmail.com",
    description="Automatic localization for various URIs (s3://, gs://, http://, https:// and local path)",
    long_description="https://github.com/ENCODE-DCC/autouri",
    long_description_content_type="text/markdown",
    url="https://github.com/ENCODE-DCC/autouri",
    packages=setuptools.find_packages(exclude=["docs", "test"]),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
    ],
    install_requires=[
        "requests",
        "pyopenssl",
        "google-cloud-storage",
        "boto3",
        "awscli",
        "dateparser",
        "filelock",
        "six>=1.13.0",
    ],
)



import hashlib
import os
import time
from typing import Any, Tuple

import pytest

from autouri.autouri import AutoURI, URIBase
from autouri.gcsuri import GCSURI
from autouri.httpurl import HTTPURL, ReadOnlyStorageError

from .files import common_paths, v6_txt_contents


@pytest.mark.parametrize("path", common_paths())
def test_httpurl_uri(path) -> Any:
    assert HTTPURL(path).uri == path


@pytest.mark.parametrize("path", common_paths())
def test_httpurl_uri_wo_ext(path) -> str:
    assert HTTPURL(path).uri_wo_ext == os.path.splitext(path.split("?", 1)[0])[0]


@pytest.mark.parametrize("path", common_paths())
def test_httpurl_uri_wo_scheme(path) -> str:
    assert HTTPURL(path).uri_wo_scheme == path.replace("https://", "", 1).replace(
        "http://", "", 1
    )


@pytest.mark.parametrize("path", common_paths())
def test_httpurl_is_valid(path) -> bool:
    """Also tests AutoURI auto-conversion since it's based on is_valid property
    """
    expected = path.startswith(("https://", "http://"))
    assert HTTPURL(path).is_valid == expected
    assert not expected or type(AutoURI(path)) == HTTPURL


@pytest.mark.parametrize("path", common_paths())
def test_httpurl_dirname(path) -> str:
    assert HTTPURL(path).dirname == os.path.dirname(path)


@pytest.mark.parametrize("path", common_paths())
def test_httpurl_dirname_wo_scheme(path) -> str:
    assert HTTPURL(path).dirname_wo_scheme == os.path.dirname(path).replace(
        "https://", "", 1
    ).replace("http://", "", 1)


@pytest.mark.parametrize("path", common_paths())
def test_httpurl_loc_dirname(path) -> str:
    assert HTTPURL(path).loc_dirname == hashlib.md5(path.encode("utf-8")).hexdigest()


@pytest.mark.parametrize("path", common_paths())
def test_httpurl_basename(path) -> str:
    assert HTTPURL(path).basename == os.path.basename(path).split("?", 1)[0]


@pytest.mark.parametrize("path", common_paths())
def test_httpurl_basename_wo_ext(path) -> str:
    assert (
        HTTPURL(path).basename_wo_ext
        == os.path.splitext(os.path.basename(path).split("?", 1)[0])[0]
    )


@pytest.mark.parametrize("path", common_paths())
def test_httpurl_ext(path) -> str:
    assert HTTPURL(path).ext == os.path.splitext(path.split("?", 1)[0])[1]


def test_httpurl_exists(url_v6_txt):
    assert HTTPURL(url_v6_txt).exists
    assert not HTTPURL(url_v6_txt + ".should-not-be-here").exists
    assert not HTTPURL("http://hey.this.should/not/be/here.txt").exists


def test_httpurl_mtime(gcs_v6_txt, url_v6_txt):
    """Write on GCS bucket which is open to public.
    Access to that file via URL.

    For URLs hosted on GCS, "last-modified" key in the header doesn't seem to be
    accurate for the following cases:
        - After writing on an existing bucket item. GCS object has a correct "update" time
        but corresponding URL's "last-modified" still points to creation time.
    """
    return

    u = GCSURI(gcs_v6_txt + ".tmp")
    if u.exists:
        # if we don't delete it, "last-modified" will be inaccurate (pointing ot creation time).
        # see the above comments in this function's docstring.
        u.rm()
    u.write("temp file for testing")
    u_url = HTTPURL(url_v6_txt + ".tmp")
    now = time.time()
    assert now - 10 < u_url.mtime < now + 10
    u.rm()
    assert not u.exists


def test_httpurl_size(url_v6_txt, v6_txt_size):
    assert HTTPURL(url_v6_txt).size == v6_txt_size


def test_httpurl_md5(url_v6_txt, v6_txt_md5_hash):
    assert HTTPURL(url_v6_txt).md5 == v6_txt_md5_hash


def test_httpurl_md5_from_file(gcs_v6_txt_url, url_v6_txt, v6_txt_md5_hash):
    u_gcs_md5 = GCSURI(gcs_v6_txt_url + URIBase.MD5_FILE_EXT)
    if u_gcs_md5.exists:
        u_gcs_md5.rm()
    u_md5 = HTTPURL(url_v6_txt + URIBase.MD5_FILE_EXT)

    assert not u_gcs_md5.exists
    assert not u_md5.exists
    u = HTTPURL(url_v6_txt)
    assert u.md5_from_file is None

    u.get_metadata(make_md5_file=True)
    # HTTPURL should not make md5 file even with make_md5_file=True
    assert not u_md5.exists
    assert u.md5_from_file is None


def test_httpurl_md5_file_uri(url_v6_txt):
    assert (
        HTTPURL(url_v6_txt + URIBase.MD5_FILE_EXT).uri
        == url_v6_txt + URIBase.MD5_FILE_EXT
    )


def test_httpurl_cp_url(url_v6_txt, url_test_path) -> "AutoURI":
    """Test copying local_v6_txt to the following destination storages:
        url_test_path: url -> url
            This will fail as intended since URL is read-only.
    """
    u = HTTPURL(url_v6_txt)
    basename = os.path.basename(url_v6_txt)

    for test_path in (url_test_path,):
        u_dest = AutoURI(os.path.join(test_path, "test_httpurl_cp", basename))

        with pytest.raises(ReadOnlyStorageError):
            _, ret = u.cp(u_dest, return_flag=True)


def test_httpurl_cp(
    url_v6_txt, local_test_path, s3_test_path, gcs_test_path
) -> "AutoURI":
    """Test copying local_v6_txt to the following destination storages:
        local_test_path: url -> local
        s3_test_path: url -> s3
        gcs_test_path: url -> gcs

    Parameters to be tested:
        no_lock:
            Copy with no locking mechanism. There is no way to test this thoroughly here.
            This will be tested with multiple threads later in test_rece_cond.py.
        no_checksum:
            Don't check md5-hash/size/mtime to skip copying (even if file already exists on destination).
        make_md5_file:
            Make md5 file on destination only when it's REQUIRED.
            It's required only if we need to compare md5 hash of source and target.
    """
    u = HTTPURL(url_v6_txt)
    basename = os.path.basename(url_v6_txt)

    for test_path in (local_test_path, s3_test_path, gcs_test_path):
        u_dest = AutoURI(os.path.join(test_path, "test_httpurl_cp", basename))
        if u_dest.exists:
            u_dest.rm()

        assert not u_dest.exists
        _, ret = u.cp(u_dest, return_flag=True)
        assert u_dest.exists and u.read() == u_dest.read() and ret == 0
        u_dest.rm()

        assert not u_dest.exists
        # cp without lock will be tested throughly in test_race_cond.py
        _, ret = u.cp(u_dest, no_lock=True, return_flag=True)
        assert u_dest.exists and u.read() == u_dest.read() and ret == 0
        u_dest.rm()

        # trivial: copy without checksum when target doesn't exists
        assert not u_dest.exists
        _, ret = u.cp(u_dest, no_checksum=True, return_flag=True)
        assert u_dest.exists and u.read() == u_dest.read() and ret == 0

        # copy without checksum when target exists
        m_dest = u_dest.get_metadata()
        assert m_dest.exists
        time.sleep(1)
        _, ret = u.cp(u_dest, no_checksum=True, return_flag=True)
        # compare new mtime vs old mtime
        # new time should be larger if it's overwritten as intended
        assert u_dest.mtime > m_dest.mtime and u.read() == u_dest.read() and ret == 0

        # copy with checksum when target exists
        m_dest = u_dest.get_metadata()
        assert m_dest.exists
        _, ret = u.cp(u_dest, return_flag=True)
        # compare new mtime vs old mtime
        # new time should be the same as old time
        assert u_dest.mtime == m_dest.mtime and u.read() == u_dest.read() and ret == 1

        # make_md5_file works only when it's required
        # i.e. when we need to compare md5 hash of src vs target
        # so target must exist prior to test it
        assert u_dest.exists
        # delete md5 file if exists
        u_dest_md5_file = AutoURI(u_dest.uri + URIBase.MD5_FILE_EXT)
        if u_dest_md5_file.exists:
            u_dest_md5_file.rm()
        _, ret = u.cp(u_dest, make_md5_file=True, return_flag=True)
        assert u_dest.exists and u.read() == u_dest.read() and ret == 1
        u_dest.rm()


def test_httpurl_write(url_test_path):
    u = HTTPURL(url_test_path + "/test_httpurl_write.tmp")
    with pytest.raises(ReadOnlyStorageError):
        u.write("test")


def test_httpurl_rm(url_test_path):
    u = HTTPURL(url_test_path + "/test_httpurl_rm.tmp")
    with pytest.raises(ReadOnlyStorageError):
        u.rm()


def test_httpurl_get_metadata(url_v6_txt, v6_txt_size, v6_txt_md5_hash):
    u = HTTPURL(url_v6_txt)

    m1 = u.get_metadata()
    assert m1.md5 == v6_txt_md5_hash
    assert m1.size == v6_txt_size

    m2 = u.get_metadata(skip_md5=True)
    assert m2.md5 is None
    assert m2.size == v6_txt_size

    u_md5 = HTTPURL(url_v6_txt + ".md5")
    if u_md5.exists:
        u_md5.rm()
    m3 = u.get_metadata(make_md5_file=True)
    assert m3.md5 == v6_txt_md5_hash
    assert m3.size == v6_txt_size
    # HTTPURL should not make md5 file even with make_md5_file=True
    assert not u_md5.exists


def test_httpurl_read(url_v6_txt):
    u = HTTPURL(url_v6_txt)
    assert u.read() == v6_txt_contents()
    assert u.read(byte=True) == v6_txt_contents().encode()


def test_httpurl_find_all_files(url_test_path):
    prefix = os.path.join(url_test_path, "test_httpurl_find_all_files")

    with pytest.raises(NotImplementedError):
        HTTPURL(prefix).find_all_files()


def test_httpurl_rmdir(url_test_path):
    prefix = os.path.join(url_test_path, "test_httpurl_rmdir")

    with pytest.raises(NotImplementedError):
        HTTPURL(prefix).rmdir()


# original methods in HTTPURL


# classmethods
def test_httpurl_get_path_sep() -> str:
    assert HTTPURL.get_path_sep() == os.path.sep


def test_httpurl_get_schemes() -> Tuple[str, ...]:
    assert HTTPURL.get_schemes() == ("http://", "https://")


def test_httpurl_get_loc_suffix() -> str:
    assert HTTPURL.get_loc_suffix() == ".url"


def test_httpurl_get_loc_prefix() -> str:
    """HTTPURL is read-only storage and hence loc_prefix == ''
    """
    assert HTTPURL.get_loc_prefix() == ""


def test_httpurl_localize(
    url_test_path, gcs_j1_json, gcs_v41_json, gcs_v421_tsv, gcs_v5_csv, gcs_v6_txt
) -> Tuple[str, bool]:
    """Localize should fail.
    """
    loc_prefix = os.path.join(url_test_path, "test_httpurl_localize")

    for j1_json in (gcs_j1_json,):
        # localization from local storage
        u_j1_json = AutoURI(j1_json)
        loc_prefix_ = loc_prefix + u_j1_json.__class__.get_loc_suffix()

        # for localization both with or without recursive
        # nothing should be localized actually
        # since they are already on a local storage
        # so loc_prefix directory itself shouldn't be created
        with pytest.raises(ReadOnlyStorageError):
            loc_uri, localized = HTTPURL.localize(
                u_j1_json, recursive=False, loc_prefix=loc_prefix_
            )
